use std::io;
use crate::identifier::Identifier;
use crate::packet::raw_packet::RawPacket;
use crate::packet::{Packet, ToPacket};
use crate::packet::packet_data::PacketData;

pub mod vec;
pub mod errors;

#[derive(Debug, Clone)]
pub enum NodeManagementCommand {
    Ping,
    RegisterNode { device_id: Identifier },
    RemoveNode,
    RegisterProperty {
        property_name: String,
        data_type: u8,
        read_only: bool,
    },
    RemoveProperty {
        property_name: String,
    },
    Error {
        error_code: u8,
        metadata: Vec<u8>,
    },
}

impl NodeManagementCommand {
    pub fn command(&self) -> u8 {
        match self {
            NodeManagementCommand::Ping => 0x00,
            NodeManagementCommand::RegisterNode { .. } => 0x01,
            NodeManagementCommand::RemoveNode => 0x02,
            NodeManagementCommand::RegisterProperty { .. } => 0x03,
            NodeManagementCommand::RemoveProperty { .. } => 0x04,
            NodeManagementCommand::Error { .. } => 0x0E
        }
    }
}

impl ToPacket for NodeManagementCommand {
    fn to_packet(self, src: Identifier, dest: Identifier, packet_id: Identifier, reply_to: Identifier) -> Packet {
        Packet {
            src, dest, packet_id, reply_to,
            command: self.command(),
            data: PacketData::NodeManagement(self.clone())
        }
    }
}

impl TryFrom<RawPacket> for NodeManagementCommand {
    type Error = io::Error;
    fn try_from(raw_packet: RawPacket) -> io::Result<Self> {
        let build_invalid_input_err = |r: &str| {
            Err(io::Error::new(io::ErrorKind::InvalidInput, r))
        };
        match raw_packet.command {
            0x00 => Ok(NodeManagementCommand::Ping),
            0x01 => if raw_packet.data_length == 4 {
                Ok(NodeManagementCommand::RegisterNode {
                    device_id: Identifier::from(raw_packet.data)
                })
            } else {
                build_invalid_input_err("expected 4 bytes")
            },
            0x02 => Ok(NodeManagementCommand::RemoveNode),
            0x03 => if raw_packet.data_length == 34 {
                Ok(NodeManagementCommand::RegisterProperty {
                    property_name: String::from_utf8(raw_packet.data[..32].to_vec())
                        .map_err(|_| io::Error::new(io::ErrorKind::InvalidInput, "String is not UTF-8"))?,
                    data_type: raw_packet.data[33],
                    read_only: raw_packet.data[34] != 0,
                })
            } else {
                build_invalid_input_err("expected 34 bytes")
            },
            0x04 => if raw_packet.data_length == 32 {
                Ok(NodeManagementCommand::RemoveProperty {
                    property_name: String::from_utf8(raw_packet.data.to_vec())
                        .map_err(|_| io::Error::new(io::ErrorKind::InvalidInput, "String is not UTF-8"))?
                })
            } else {
                build_invalid_input_err("expected 32 bytes")
            },
            0x0E => if raw_packet.data_length == 3 {
                Ok(NodeManagementCommand::Error {
                    error_code: raw_packet.data[0],
                    metadata: {
                        let metadata_length = u16::from_be_bytes([raw_packet.data[1], raw_packet.data[2]]) as usize;
                        raw_packet.data[3..metadata_length].to_vec()
                    },
                })
            } else {
                build_invalid_input_err("expected 3 bytes")
            },
            _ => Err(io::Error::new(io::ErrorKind::InvalidInput, "command group is unsupported"))
        }
    }
}
