use std::io;

use crate::{
    identifier::Identifier,
    packet::{packet_data::PacketData, raw_packet::RawPacket, Packet, ToPacket},
    prelude::as_u32_be,
};

pub mod vec;

#[derive(Debug, Clone)]
pub enum RawDataTransmission {
    SetupTransmission {
        /// The total size of the data being sent.
        size: u64,
        /// This string is the mime type of the data being sent.  
        /// By default this should be `application/octet-stream`.  
        /// The maximum length of this string is 128 bytes since the mime rfc states the recommended max is 127 bytes.
        mime_type: String,
    },
    Data {
        /// The number of the segment in the sequence of segments being sent.
        sequence_number: u32,
        /// Segment size in bytes
        size: u32,
        /// A segment of the data being sent
        data: Vec<u8>,
    },
}

impl ToPacket for RawDataTransmission {
    fn to_packet(
        self,
        src: Identifier,
        dest: Identifier,
        packet_id: Identifier,
        reply_to: Identifier,
    ) -> Packet {
        Packet {
            src,
            dest,
            packet_id,
            reply_to,
            command: 0x20,
            data: PacketData::RawDataTransmission(self.clone()),
        }
    }
}

impl TryFrom<RawPacket> for RawDataTransmission {
    type Error = io::Error;
    fn try_from(raw_packet: RawPacket) -> io::Result<Self> {
        let build_invalid_input_err = |r: &str| Err(io::Error::new(io::ErrorKind::InvalidInput, r));

        let mut data = raw_packet.data;
        match raw_packet.command {
            0x20 => {
                let size = u64::from_be_bytes([
                    data[0], data[1], data[2], data[3], data[4], data[5], data[6], data[7],
                ]);
                let mime_type = String::from_utf8(data[8..].to_vec()).map_err(|_| {
                    io::Error::new(io::ErrorKind::InvalidInput, "Invalid mime type string")
                })?;
                Ok(RawDataTransmission::SetupTransmission { size, mime_type })
            }
            0x21 => {
                let sequence_number = as_u32_be(&mut data);
                let size = as_u32_be(&mut data);
                Ok(RawDataTransmission::Data {
                    sequence_number,
                    size,
                    data,
                })
            }
            _ => build_invalid_input_err("Invalid command"),
        }
    }
}
