use std::io;
use crate::data_types::{DataType, get_data_types};
use crate::identifier::{Identifier, self};
use crate::prelude::as_u64_be;

impl Into<Vec<u8>> for DataType {
    fn into(self) -> Vec<u8> {
        match self {
            DataType::Nothing => vec![0x00],
            DataType::Array(vec) => {
                let mut res = vec![0x01];
                for item in vec {
                    res.extend(Into::<Vec<u8>>::into(item).iter());
                }
                res
            },
            DataType::Boolean(b) => vec![0x10, b as u8],
            DataType::Number(v) => {
                let mut bytes = vec![0x11];
                bytes.extend(v.to_be_bytes());
                bytes
            },
            DataType::Text(s) => {
                let mut bytes = vec![0x12];
                bytes.extend(s.into_bytes());
                bytes
            },
            DataType::Identifier(identifier) => {
                let mut bytes = vec![0x13];
                bytes.extend(identifier.bytes);
                bytes
            },
            DataType::RGB(r, g, b) => vec![0x20, r, g, b],
        }
    }
}

macro_rules! impl_data_type {
    ($v: ident, $len: literal, $($t:tt)*) => {
        if $v.len() - 1 == $len {
            Ok($($t)*)
        } else {
            Err(io::Error::new(io::ErrorKind::InvalidData, format!("Expected data length {}, actual {}", $len, $v.len())))
        }
    };
}

/// Turn a `Vec<u8>` into a single data type.
/// Returns Datatype::Nothing if it's nothing or can find nothing.
impl TryFrom<Vec<u8>> for DataType {
    type Error = io::Error;
    fn try_from(value: Vec<u8>) -> io::Result<Self> {
        if value.len() == 0 {
            return Err(io::Error::new(io::ErrorKind::InvalidData, "data length == 0"))
        };
        match value[0] {
            0x01 => {
                if value.len() - 1 < 2 { return Err(io::Error::new(io::ErrorKind::InvalidInput, "The array DataType requires a 2 byte header.")) }
                let size = ((value[1] as u16) << 8) | value[2] as u16;
                if value.len() - 1 < size as usize { return Err(io::Error::new(io::ErrorKind::InvalidData, "Array specified data length != array data length")) }
                Ok(DataType::Array(get_data_types(value[3..(size as usize)].to_vec())))
            },
            0x10 => impl_data_type!(value, 1, DataType::Boolean(value[1] != 0)),
            0x11 => impl_data_type!(value, 8, DataType::Number(u64::from_be_bytes([
                value[1], value[2], value[3], value[4],
                value[5], value[6], value[7], value[8]
            ]))),
            0x12 => impl_data_type!(
                value,
                256,
                if let Ok(s) = String::from_utf8(value[1..257].to_vec()) {
                    DataType::Text(s)
                } else {
                    DataType::Nothing
                }
            ),
            0x13 => impl_data_type!(value, 4, DataType::Identifier(Identifier::from(&value[1..5]))),
            0x20 => impl_data_type!(value, 3, DataType::RGB(value[1], value[2], value[3])),
            _ => Err(io::Error::new(io::ErrorKind::InvalidData, "Could not match data type header to a data type.")),
        }
    }
}
